@@ -31,12 +31,13 @@
 
 #include "glibtop_private.h"
 
+#include "procmap_smaps.c"
 
 #define  MAPS_FILE "/proc/%u/maps"
 #define SMAPS_FILE "/proc/%u/smaps"
 
 
-#define PROC_MAPS_FORMAT "%16llx-%16llx %4c %16llx %02hx:%02hx %llu%*[ ]%n"
+#define PROC_MAPS_FORMAT "%16" G_GINT64_MODIFIER "x-%16" G_GINT64_MODIFIER "x %4c %16" G_GINT64_MODIFIER "x %02hx:%02hx %" G_GINT64_MODIFIER "u%*[ ]%n"
 
 
 static const unsigned long _glibtop_sysdeps_proc_map =
@@ -66,35 +67,19 @@ _glibtop_init_proc_map_s (glibtop *server)
 /* Provides detailed information about a process. */
 
 
-
-struct smap_value {
-	char name[16];
-	size_t name_len;
-	ptrdiff_t offset;
-};
-
-static int
-compare(const void* a_key, const void* a_smap)
-{
-	const char* key = a_key;
-	const struct smap_value* smap = a_smap;
-	return strncmp(key, smap->name, smap->name_len);
-}
-
-
-static gboolean
+static const char*
 is_smap_value(const char* s)
 {
 	for ( ; *s; ++s) {
 
 		if (isspace(*s))
-			return FALSE;
+			return NULL;
 
 		if (*s == ':')
-			return TRUE;
+			return s;
 	}
 
-	return FALSE;
+	return NULL;
 }
 
 
@@ -105,25 +90,18 @@ is_smap_value(const char* s)
 static gboolean
 parse_smaps(glibtop_map_entry *entry, const char* line)
 {
-#define SMAP_OFFSET(MEMBER) offsetof(glibtop_map_entry, MEMBER)
-#define STR_AND_LEN(X) (X), (sizeof X - 1)
 
-	/* keep sorted */
-	const struct smap_value values[] = {
-		{ STR_AND_LEN("Private_Clean:"),	SMAP_OFFSET(private_clean) },
-		{ STR_AND_LEN("Private_Dirty:"),	SMAP_OFFSET(private_dirty) },
-		{ STR_AND_LEN("Rss:"),			SMAP_OFFSET(rss) },
-		{ STR_AND_LEN("Shared_Clean:"),		SMAP_OFFSET(shared_clean) },
-		{ STR_AND_LEN("Shared_Dirty:"),		SMAP_OFFSET(shared_dirty) },
-		{ STR_AND_LEN("Size:"),			SMAP_OFFSET(size) }
-	};
+	const struct smap_value* smap;
+	size_t len;
+	const char* colon;
 
-#undef STR_AND_LEN
-#undef SMAP_OFFSET
+	if ((colon = is_smap_value(line)) == NULL)
+		return FALSE;
 
-	struct smap_value* smap;
+	len = colon - line + 1;
+	smap = _glibtop_find_smap(line, len);
 
-	smap = bsearch(line, values, G_N_ELEMENTS(values), sizeof values[0], compare);
+//		g_debug("smap %s -> %p", line, smap);
 
 	if (smap) {
 		char *offset;
@@ -133,14 +111,75 @@ parse_smaps(glibtop_map_entry *entry, const char* line)
 		offset += smap->offset;
 		value = (void*) offset;
 
-		*value = get_scaled(line + smap->name_len, NULL);
-		return TRUE;
+		*value = get_scaled(line + len, NULL);
 	}
 
-	return is_smap_value(line);
+	return TRUE;
 }
 
 
+/*
+  sscanf is too slow
+  and system-monitor calls procmap for each pid every second
+
+  manual parsing is faster
+
+  error checking is weaker
+*/
+
+static gboolean
+parse_line(char* line,
+	   guint64* start, guint64* end, char flags[4], guint64* offset,
+	   gushort* dev_major, gushort* dev_minor, guint64* inode,
+	   char** filename)
+{
+	/* %16llx-%16llx %4c %16llx %02hx:%02hx %llu%*[ ]%n */
+
+	char *p, *next;
+
+	p = line;
+
+	*start = strtoull(p, &p, 16);
+
+	if (G_UNLIKELY(*p != '-'))
+		return FALSE;
+	p++;
+
+	*end = strtoull(p, &p, 16);
+
+	p = next_token(p);
+
+	memcpy(flags, p, 4);
+	p += 4;
+
+	*offset = strtoull(p, &p, 16);
+
+	*dev_major = strtoul(p, &p, 16);
+
+	if (G_UNLIKELY(*p != ':'))
+		return FALSE;
+	p++;
+
+	*dev_minor = strtoul(p, &p, 16);
+
+	*inode = strtoull(p, &p, 10);
+
+	p = next_token(p);
+
+	*filename = p;
+
+	for ( ; *p; p++) {
+		if (*p == '\n') {
+			*p = '\0';
+			break;
+		}
+	}
+
+	return TRUE;
+}
+
+
+
 
 
 glibtop_map_entry *
@@ -153,6 +192,7 @@ glibtop_get_proc_map_s (glibtop *server, glibtop_proc_map *buf,	pid_t pid)
 	  It's the average number of entry per process on my laptop
 	*/
 
+	size_t added = 0;
 	GArray *entry_list = g_array_sized_new(FALSE, FALSE,
 					       sizeof(glibtop_map_entry),
 					       100);
@@ -180,8 +220,7 @@ glibtop_get_proc_map_s (glibtop *server, glibtop_proc_map *buf,	pid_t pid)
 	while(TRUE)
 	{
 		unsigned long perm;
-		guint len;
-		int line_end;
+		/* int line_end; */
 
 		unsigned short dev_major, dev_minor;
 		guint64 start, end, offset, inode;
@@ -195,6 +234,12 @@ glibtop_get_proc_map_s (glibtop *server, glibtop_proc_map *buf,	pid_t pid)
 
 	new_entry_line:
 
+		if (!parse_line(line,
+				&start, &end, flags, &offset,
+				&dev_major, &dev_minor, &inode, &filename))
+			continue;
+
+		/*
 		if (sscanf(line, PROC_MAPS_FORMAT,
 			   &start, &end, flags, &offset,
 			   &dev_major, &dev_minor, &inode, &line_end) != 7)
@@ -202,6 +247,7 @@ glibtop_get_proc_map_s (glibtop *server, glibtop_proc_map *buf,	pid_t pid)
 
 		filename = line + line_end;
 		g_strstrip(filename);
+		*/
 
 		/* Compute access permissions. */
 		perm = 0;
@@ -224,9 +270,12 @@ glibtop_get_proc_map_s (glibtop *server, glibtop_proc_map *buf,	pid_t pid)
 		  avoid copying the entry, grow by 1 and point to the last
 		  element.
 		*/
-		len = entry_list->len;
-		g_array_set_size(entry_list, len + 1);
-		entry = &g_array_index(entry_list, glibtop_map_entry, len);
+
+		if (G_UNLIKELY(added >= entry_list->len)) {
+			g_array_set_size(entry_list, 2 * entry_list->len);
+		}
+
+		entry = &g_array_index(entry_list, glibtop_map_entry, added++);
 
 		entry->flags = _glibtop_sysdeps_map_entry;
 		entry->start = start;
@@ -253,6 +302,7 @@ glibtop_get_proc_map_s (glibtop *server, glibtop_proc_map *buf,	pid_t pid)
 
 eof:
 
+	g_array_set_size(entry_list, added);
 	free(line);
 	fclose (maps);
 
